#pragma once

#include <algorithm>
#include <string>

#include <amdis/linear_algebra/RunnerInterface.hpp>
#include <amdis/linear_algebra/SolverInfo.hpp>
#include <amdis/linear_algebra/eigen/SolverConfig.hpp>

namespace AMDiS
{
  /**
   * \ingroup Solver
   * \class AMDiS::DirectRunner
   * \brief \implements RunnerInterface for the (external) direct solvers
   */
  template <class LU, class Matrix, class VectorX, class VectorB>
  class DirectRunner
      : public RunnerInterface<Matrix, VectorX, VectorB>
  {
  protected:
    using Super = RunnerInterface<Matrix, VectorX, VectorB>;

  public:
    /// Constructor.
    DirectRunner(std::string const& prefix)
      : solver_{}
    {
      SolverConfig<LU>::init(prefix, solver_);
    }

    /// Implementes \ref RunnerInterface::init()
    virtual void init(Matrix const& A) override
    {
      solver_.compute(A);

      test_exit(solver_.info() == Eigen::Success,
        "Error in solver.compute(matrix)");
    }


    /// Implementes \ref RunnerInterface::exit()
    virtual void exit() override {}

    /// Implementes \ref RunnerInterface::solve()
    virtual int solve(Matrix const& A, VectorX& x, VectorB const& b,
                      SolverInfo& solverInfo) override
    {
      x = solver_.solve(b);

      auto r = VectorB(b);
      if (x.norm() != 0)
        r -= A * x;

      solverInfo.setAbsResidual(r.norm());
      solverInfo.setError(solver_.info());

      return solver_.info() == Eigen::Success ? 0 : 1;
    }

  private:
    LU solver_;
  };
}
